/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*****************************************************************************
* Tabs 5 9
* The carets in the following lines will match up if tabs are set properly
*   ^   ^   ^   ^   ^   ^   <== tabs
*   ^   ^   ^   ^   ^   ^   <== spaces
******************************************************************************
*
*   Copyright (C) 2007 Systemic Realtime Design, LLC.    All Rights Reserved.
*
*   This source code may contain confidential information of
*            Systemic Realtime Design, LLC.   (SRD)
*   Any use, reproduction, modification, disclosure, distribution or
*   transfer of this software, or any software which includes or is based
*   upon any of this code, is prohibited unless expressly authorized
*   by SRD by written agreement.
*
*   For more information please email licensing@SysRTime.com.
*
*****************************************************************************/

#ifndef TIPC_SKBUFF_H
#define TIPC_SKBUFF_H 1

#include <stdio.h>
#include <inttypes.h>
#include <assert.h>

#include "sys/param_bsd.h"
#include "sys/mbuf.h"
#include "sys/socketvar.h"

/*
 * TIPC message buffer code
 *
 * TIPC message buffer headroom reserves space for a link-level header
 * (in case the message is sent off-node).
 *
 * Headroom is currently set to 24 bytes to facilitate support of
 * the following header formats:
 *
 * - Ethernet header (14 bytes)
 * - Ethernet header + VLAN info (14 + 4 bytes)
 * - Ethernet header + gianfar FCB info (14 + 8 bytes)
 *
 * Note: Headroom is always a multiple of 4 to ensure the TIPC header fields
 *       are word aligned for quicker access
 */

#define BUF_HEADROOM 24u

struct tipc_skb_cb {
    void *handle;
};


/*
 *  QNX TIPC emulation of common file socket buffer API
 */

typedef struct mbuf mbuf_t;
struct sk_buff;

struct sk_buff_head {
	struct sk_buff *next;
	struct sk_buff *prev;

	uint32_t qlen;
	spinlock_t lock;
};

struct sk_buff {
	struct sk_buff *next;         /* ptr to next buffer in list */
	struct sk_buff *prev;         /* ptr to previous buffer in list */
	struct sk_buff_head *list;
	mbuf_t         *mBufPtr;      /* ptr to mBuf */
	char           cb[sizeof(struct tipc_skb_cb)]; /* control block area */
	unsigned int   len;
	unsigned char  *data;
	unsigned char  *tail;
};


extern struct sk_buff *skb_copy           (const struct sk_buff *skb, int gfp_mask);
extern struct sk_buff *skb_clone          (struct sk_buff *skb, int gfp_mask);
extern int             skb_cow            (struct sk_buff *skb, unsigned int headroom);
extern void            skb_reserve        (struct sk_buff *skb, unsigned int len );
extern unsigned char  *skb_put            (struct sk_buff *skb, unsigned int len );
extern unsigned char  *skb_push           (struct sk_buff *skb, unsigned int len );
extern unsigned char  *skb_pull           (struct sk_buff *skb, unsigned int len );
extern void            skb_trim           (struct sk_buff *skb, unsigned int len);
extern struct sk_buff *alloc_skb          (unsigned int size, int gfp_mask);
extern void            kfree_skb          (struct sk_buff *skb);
extern int             skb_headroom       (const struct sk_buff *skb);
extern int             skb_tailroom       (const struct sk_buff *skb);
extern void           *skb_header_pointer (const struct sk_buff *skb, int offset, int len,
 void *buffer);
extern int             pskb_may_pull      (struct sk_buff *skb, unsigned int len);
extern int             pskb_trim          (struct sk_buff *skb, unsigned int len);
extern int             skb_cloned         (const struct sk_buff *skb);
struct sk_buff        *vxskb_attach       (mbuf_t *mBlkChain);

uint32_t        skb_queue_len      (struct sockbuf *sb);
struct sk_buff *__skb_dequeue      (struct sockbuf *sb);
struct mbuf    *skb_peek           (struct sockbuf *sb);
int             skb_queue_empty    (struct sockbuf *sb);
int             skb_queue_init     (struct socket *so, int tx_size, int rx_size);
void            skb_queue_destroy  (struct socket *so);


static inline struct sk_buff *alloc_skb_fclone(unsigned int size, int gfp_mask)
{
	return alloc_skb(size, gfp_mask);
}


#define TIPC_SKB_CB(__skb) ((struct tipc_skb_cb *)&((__skb)->cb[0]))

static inline struct tipc_msg *buf_msg(struct sk_buff *skb)
{
    return (struct tipc_msg *)skb->data;
}


/* from vxworks port of skbuff */
#define vxskb_clsize(skb) (skb->mBufPtr->m_ext.ext_size)
#define vxskb_head(skb)   (uint8_t*)(skb->mBufPtr->m_ext.ext_buf)
#define vxskb_end(skb)    (uint8_t*)(skb->mBufPtr->m_ext.ext_buf \
							+ vxskb_clsize(skb))

/* buffer macros (allows buf debugging) */
#define MBUF_FREE(_BUF_)		{ m_freem(_BUF_); _BUF_=NULL; }
#define MBUF_COPY(_BUF_...)		m_dup(_BUF_)
#define MBUF_DUP(_BUF_...)		m_dup(_BUF_)

#endif

