/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*
 * net/tipc/tipc_socket.c: TIPC socket API
 * 
 * Copyright (c) 2001-2006, Ericsson AB
 * Copyright (c) 2004-2007, Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "tipc_core.h"
#include "tipc_socket.h"
#include "tipc_dbg.h"
#include "tipc_skbuff.h"

#include "sys/param_bsd.h"
#include "sys/mbuf.h"

#include <sys/types.h>
#include <sys/neutrino.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/socketvar.h>

#define PRUREQUESTS
#include <sys/protosw.h>

#include <sys/domain.h>
#include <sys/iofunc.h>
#include <sys/dispatch.h>
#include <sys/proc.h>
#include <fcntl.h>
#include "nw_msg.h"
#include "nw_datastruct.h"

#include <linux/tipc.h>
#include <linux/tipc_config.h>
#include <net/tipc/tipc_plugin_msg.h>
#include <net/tipc/tipc_plugin_port.h>


#ifdef CONFIG_TIPC_SOCKET_API


typedef enum {
    SS_FREE = 0,            /* not allocated               */
    SS_UNCONNECTED,         /* unconnected to any socket   */
    SS_CONNECTING,          /* in process of connecting    */
    SS_CONNECTED,           /* connected to socket         */
    SS_DISCONNECTING,       /* in process of disconnecting */
	SS_LISTENING = -1,      /* socket is listening         */
	SS_READY = -2,          /* socket is connectionless    */
} socket_state;

#define sock_state(sk)  ((sk)->state)
#define sock_port(sk)   ((struct tipc_port *)(sk)->sk_user_data)

#define OVERLOAD_LIMIT_BASE     5000
#define CONN_TIMEOUT_DEFAULT    8000    /* default connect() timeout (in ms) */

#define SET_SOCK_ADDR 0
#define SET_PEER_ADDR 1

static int backlog_rcv (struct sock *sk, struct sk_buff *skb);
static u32 dispatch (struct tipc_port *tport, struct sk_buff *buf);
static void wakeupdispatch (struct tipc_port *tport);
static int tipc_sock_get_name(struct socket *sock, struct sockaddr_tipc *addr,
				    		  int *uaddr_len, int peer);
int tipc_sock_send_msg    (struct socket *sock, struct msghdr *m, int total_len, void *iocb);
int tipc_sock_send_packet (struct socket *sock, struct msghdr *m, int total_len, void *iocb);
int tipc_sock_accept (struct socket *new_sock, struct sk_buff *buf, struct mbuf *mb);

static int sockets_enabled = 0;
static atomic_t tipc_queue_size = ATOMIC_INIT(0);

#define sock_mutex_enter(sk)    {pthread_mutex_lock(&sk->sk_lock);}
#define sock_mutex_exit(sk)     {pthread_mutex_unlock(&sk->sk_lock);}
static inline int sock_mutex_owned(struct sock *sk)
{
	int ret = pthread_mutex_trylock (&sk->sk_lock);
	if (ret == EBUSY)
		return 1;
	pthread_mutex_unlock(&sk->sk_lock);
	return 0;
}


static inline int sock_wait (struct sock *sk, uint32_t ms_timeout)
{
	int ret;
	uint32_t val = ms_timeout / (1000000 / NTO_mHZ);

	/* mutex must be owned */
	if (!sock_mutex_owned(sk))
		return EINVAL;

	pthread_mutex_unlock(&sk->sk_lock);
	ret = tsleep ((caddr_t)&sk->sk_sleep, PUSER, "sk_sleep", val);
	pthread_mutex_lock(&sk->sk_lock);
	return ret;
}

static inline void sock_wakeup (struct sock *sk)
{
	wakeup((caddr_t)&sk->sk_sleep);
}


static void handle_backlog (struct sock *sk)
{
	struct sk_buff *skb = sk->sk_backlog.head;
	struct sk_buff *next;

	ASSERT(sock_mutex_owned(sk));

	while (skb != NULL) {
		next = skb->next;
		skb->next = NULL;
		sk->sk_backlog_rcv(sk, skb);
		skb = next;
	}
	sk->sk_backlog.head = sk->sk_backlog.tail = NULL;
}


/*
 * lock_sock()
 *
 * Locks the socket lock
 */
static void lock_sock(struct sock *sk)
{
	int owned = sock_mutex_owned(sk);

	PRINT_DBG("%s: owned: %d\n", __FUNCTION__, owned);	
	if (!owned)
		sock_mutex_enter(sk);

	sk->waiters++;
	PRINT_DBG("%s: waiters %d, owner: 0x%x\n", __FUNCTION__, sk->waiters, sk->owner);	
	while (sk->owner != 0)
		sock_wait(sk, 0);
	sk->owner = pthread_self();
	sk->waiters--;

	if (!owned)
		sock_mutex_exit(sk);
	
	PRINT_DBG("%s: waiters: %d\n", __FUNCTION__, sk->waiters);
}


/*
 * release_sock()
 *
 * Releases the socket lock
 */
static void release_sock(struct sock *sk)
{
	int owned = sock_mutex_owned(sk);

	PRINT_DBG("%s: owned: %d\n", __FUNCTION__, owned);	
	if (!owned)
		sock_mutex_enter(sk);
	ASSERT(sk->owner == pthread_self());

	/* Process all messages that arrived while we owned socket */	
	handle_backlog(sk);
	sk->owner = 0;
	if (sk->waiters)	
		sock_wakeup(sk);

	if (!owned)
		sock_mutex_exit(sk);
	PRINT_DBG("%s: released\n", __FUNCTION__);
}


/*
 * rx_queue_full - determine if receive queue can accept another message
 * @msg: message to be added to queue
 * @queue_size: current size of queue
 * @base: nominal maximum size of queue
 *
 * Returns 1 if queue is unable to accept message, 0 otherwise
 */
static int rx_queue_full(struct tipc_msg *msg, u32 queue_size, u32 base)
{
	u32 threshold;
	u32 imp = msg_importance(msg);

	if (imp == TIPC_LOW_IMPORTANCE)
		threshold = base;
	else if (imp == TIPC_MEDIUM_IMPORTANCE)
		threshold = base * 2;
	else if (imp == TIPC_HIGH_IMPORTANCE)
		threshold = base * 100;
	else
		return (0);

	if (msg_connected(msg))
		threshold *= 4;

	return (queue_size >= threshold);
}


/*
 * filter_rcv - validate incoming message
 * @sk: socket
 * @buf: message
 *
 * Enqueues message on receive queue if acceptable; optionally handles
 * disconnect indication for a connected socket.
 *
 * Called with socket lock already taken; port lock may also be taken.
 *
 * Returns TIPC error status code (TIPC_OK if message is not to be rejected)
 */

static u32 filter_rcv(struct sock *sk, struct sk_buff *buf)
{
	struct socket *sock = sk->sk_socket;
	struct tipc_msg *msg = buf_msg(buf);
	u32 recv_q_len;
	
	struct mbuf *mb = buf->mBufPtr;
	struct sockaddr_tipc sa;
	int res = TIPC_OK;

	PRINT_DBG("%s: start\n", __FUNCTION__);

	ASSERT(sock_mutex_owned(sk));

	/* Reject message if it is wrong sort of message for socket */

	/*
	 * WOULD IT BE BETTER TO JUST DISCARD THESE MESSAGES INSTEAD?
	 * "NO PORT" ISN'T REALLY THE RIGHT ERROR CODE, AND THERE MAY
	 * BE SECURITY IMPLICATIONS INHERENT IN REJECTING INVALID TRAFFIC
	 */

	if (sock_state(sk) == SS_READY) {
		if (msg_connected(msg)) {
			msg_dbg(msg, "dispatch filter 1\n");
			return (TIPC_ERR_NO_PORT);
		}
	} else {
		if (msg_mcast(msg)) {
			msg_dbg(msg, "dispatch filter 2\n");
			return (TIPC_ERR_NO_PORT);
		}
		if (sock_state(sk) == SS_CONNECTED) {
			if (!msg_connected(msg)) {
				msg_dbg(msg, "dispatch filter 3\n");
				return (TIPC_ERR_NO_PORT);
			}
		} else if (sock_state(sk) == SS_CONNECTING) {
			if (!msg_connected(msg) && (msg_errcode(msg) == 0)) {
				msg_dbg(msg, "dispatch filter 4\n");
				return (TIPC_ERR_NO_PORT);
			}
		} else if (sock_state(sk) == SS_LISTENING) {
			if (msg_connected(msg) || msg_errcode(msg)) {
				msg_dbg(msg, "dispatch filter 5\n");
				return (TIPC_ERR_NO_PORT);
			}
		} else if (sock_state(sk) == SS_DISCONNECTING) {
			msg_dbg(msg, "dispatch filter 6\n");
			return (TIPC_ERR_NO_PORT);
		} else {	/* (sock_state(sk) == SS_UNCONNECTED) */
			if (msg_connected(msg) || msg_errcode(msg)) {
				msg_dbg(msg, "dispatch filter 7\n");
				return (TIPC_ERR_NO_PORT);
			}
		}
	}
	
	/* Reject message if there isn't room to queue it */

	recv_q_len = skb_queue_len(&sock->so_rcv);
	if (unlikely(recv_q_len >= (OVERLOAD_LIMIT_BASE / 2))) {
		if (rx_queue_full(msg, recv_q_len, OVERLOAD_LIMIT_BASE / 2)) {
			err ("%s: rx_queue_full (queue_len: %d)\n", __FUNCTION__, recv_q_len);
			res = TIPC_ERR_OVERLOAD;
			goto FILTER_RCV_EXIT;	
		}
	}
	PRINT_DBG("%s: message passed 2nd queue filter\n", __FUNCTION__);

	/* Initiate connection termination for an incoming 'FIN' */

	if (unlikely(msg_errcode(msg) && (sock_state(sk) == SS_CONNECTED))) {
		if (msg_errcode(msg) == TIPC_CONN_SHUTDOWN)
		{
			PRINT_DBG ("%s: connection termination msg received\n", __FUNCTION__);
			sock_state(sk) = SS_DISCONNECTING;
			tipc_disconnect_port(sock_port(sk));
			soisdisconnecting(sock);	

			MBUF_FREE(mb);
			res = TIPC_OK;	
		}
		else 
		{
			warn ("%s: error msg received! (0x%x)\n", __FUNCTION__, msg_errcode(msg));
			res = msg_errcode(msg);	
		}
		goto FILTER_RCV_EXIT;	
	}

	/* Enqueue message (finally!) */

	msg_dbg(msg, "<DISP<: ");
	TIPC_SKB_CB(buf)->handle = msg_data(msg);
	atomic_inc(&tipc_queue_size);
	
	PRINT_DBG("%s: queuing message on socket 0x%x, msg_sz %d, hdr_sz %d\n", 
		__FUNCTION__, (int)sock, msg_size(msg), msg_hdr_sz(msg)); 

	/* adjust data ptrs and len to start of data */	
	if (msg_data_sz(msg))
		m_adj(mb, msg_hdr_sz(msg));

	/* let BSD stack know that a new connection attempt was received */
	if (sock_state(sk) == SS_LISTENING)
	{
		struct socket *new_sock;
		PRINT_DBG("%s: -------- creating new socket connection from sock 0x%x -----------\n", __FUNCTION__, (int)sock);
		new_sock = sonewconn(sock, SS_ISCONFIRMING);
		if (new_sock == NULL)
		{
			err("%s: sonewconn failed: %s\n", __FUNCTION__, strerror(errno));
			MBUF_FREE(mb);
			res = TIPC_OK;
			goto FILTER_RCV_EXIT;	
		}

		/* do accept here cause PRU_ACCEPT callback has no access to tipc_msg */
		PRINT_DBG("%s: tipc_sock_accept sock: 0x%x\n", __FUNCTION__, (int)sock);
		res = tipc_sock_accept (new_sock, buf, mb);
		if (res != EOK)
		{
			err("%s: tipc_sock_accept failed: %s\n", __FUNCTION__, strerror(errno));
			MBUF_FREE(mb);	
		}
		res = TIPC_OK;
		goto FILTER_RCV_EXIT;
	}

	/* queue the mbuf depending on socket type */
	if (sock->so_type == SOCK_SEQPACKET) 
	{
		PRINT_DBG("%s: sbappendrecord (%d space in buf, need %d)\n", 
			__FUNCTION__, sbspace(&sock->so_rcv), mb->m_pkthdr.len);
		if (mb->m_pkthdr.len > sbspace(&sock->so_rcv))
		{
			err("%s: queuing message failed! Need %d bytes, %d bytes space left.\n",
				__FUNCTION__, mb->m_pkthdr.len, sbspace(&sock->so_rcv));
			
			sock->so_rcv.sb_overflowed++;
			res = TIPC_ERR_OVERLOAD;
			goto FILTER_RCV_EXIT;	
		}

		PRINT_DBG("%s: sbappendrecord (%d in buf)\n", __FUNCTION__, sock->so_rcv.sb_cc);
		sbappendrecord(&sock->so_rcv, mb);
		res = TIPC_OK;
	}
	else
	{
		/* setup senders sockaddr_tipc struct (see code from set_orig_addr function) */
		memset(&sa, 0, sizeof(struct sockaddr_tipc));	
		sa.family = AF_TIPC;
		sa.addrtype = TIPC_ADDR_ID;
		sa.addr.id.ref = msg_origport(msg);
		sa.addr.id.node = msg_orignode(msg);
		sa.addr.name.domain = 0; /* could leave uninitialized */
		sa.scope = 0;            /* could leave uninitialized */
		sa.sockaddr_sa_len = sizeof(struct sockaddr_tipc); /* sbappendaddr needs sa_len field from struct sockaddr */

		PRINT_DBG("%s: sbappendaddr (%d in buf)\n", __FUNCTION__, sock->so_rcv.sb_cc);
		res = sbappendaddr(&sock->so_rcv, (struct sockaddr*)&sa, mb, NULL);
		if (res == 0) 
		{
			sock->so_rcv.sb_overflowed++;
			err("%s: queuing message failed! No space left? (%d bytes in buf)\n", 
				__FUNCTION__, sock->so_rcv.sb_cc); 
			
			res = TIPC_ERR_OVERLOAD;
			goto FILTER_RCV_EXIT;	
		}
		res = TIPC_OK;
	}


FILTER_RCV_EXIT:
	
	/* data is ready, wakeup read blocks */
	sorwakeup(sock);
	wakeup((caddr_t) &sock->so_timeo);
	
	sk->need_pollwakeup = 1;
	sock_wakeup(sk);

	/* If res == OK, it means the mbuf was queued (or freed)
	 * detach the mbuf from the skbuff and free the skbuff
	 */
	if (res == TIPC_OK)	{
		buf->mBufPtr = NULL;
		vxskb_dealloc(buf);	
	}
	return res;
}




/**
 * backlog_rcv - handle incoming message from backlog queue
 * @sk: socket
 * @buf: message
 *
 * Caller must hold socket lock, but not port lock.
 *
 * Returns 0
 */

static int backlog_rcv(struct sock *sk, struct sk_buff *buf)
{
	u32 res;

	res = filter_rcv(sk, buf);
	if (res)
		tipc_reject_msg(buf, res);
	return 0;
}




static void sk_add_backlog(struct sock *sk, struct sk_buff *buf)
{
	ASSERT(sock_mutex_owned(sk));
	if (!sk->sk_backlog.tail) {
		sk->sk_backlog.head = sk->sk_backlog.tail = buf;
	} else {
		(sk->sk_backlog.tail)->next = buf;
		sk->sk_backlog.tail = buf;
	}
	buf->next = NULL;
}


/**
 * dispatch - handle incoming message
 * @tport: TIPC port that received message
 * @buf: message
 *
 * Called with port lock already taken.
 *
 * Returns TIPC error status code (TIPC_OK if message is not to be rejected)
 */

static u32 dispatch(struct tipc_port *tport, struct sk_buff *buf)
{
	struct sock *sk = (struct sock *)tport->usr_handle;
	u32 res;
	
	/*
	 * Process message if socket is unlocked; otherwise add to backlog
	 * queue
	 */
	sock_mutex_enter(sk); // MLL@SRD: TODO: causes lockup?????

	PRINT_DBG("%s: are we owner? %d\n", __FUNCTION__, sk->owner);
	if (sk->owner == 0) {
		lock_sock(sk);
		PRINT_DBG("%s: call filter_rcv\n", __FUNCTION__);
		res = filter_rcv(sk, buf);
		release_sock(sk);
		//ASSERT(sock_mutex_owned(sk)); // XXX: asserts on an early exit by server
	} else {
		PRINT_DBG("%s: we are owner, add to backlog\n", __FUNCTION__);
		sk_add_backlog(sk, buf);
		res = TIPC_OK;
	}
	sock_mutex_exit(sk);
	return res;
}


/**
 * wakeupdispatch - wake up port after congestion
 * @tport: port to wakeup
 *
 * Called with port lock already taken.
 */

static void wakeupdispatch(struct tipc_port *tport)
{
	struct sock *sk = (struct sock *)tport->usr_handle;
	wakeup((caddr_t)&sk->sk_sleep);
}


/*
 * auto_connect - complete connection setup to a remote port
 * @sock: socket structure
 * @msg: peer's response message
 *
 * Returns 0 on success, errno otherwise
 */
static int auto_connect(struct socket *sock, struct tipc_msg *msg)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	struct tipc_portid peer;

	if (msg_errcode(msg)) {
		sock_state(sk) = SS_DISCONNECTING;
		return (-ECONNREFUSED);
	}

    peer.ref = msg_origport(msg);
    peer.node = msg_orignode(msg);
    tipc_connect2port(tport->ref, &peer);
    tipc_set_portimportance(tport->ref, msg_importance(msg));
    sock_state(sk) = SS_CONNECTED;
			
	/* let BSD stack know that we are connected! */
	PRINT_DBG("%s: connection completed for sock 0x%x!\n", __FUNCTION__, (int)sock);
	soisconnected(sock);
    return (0);
}

/*
 * set_orig_addr - capture sender's address for received message
 * @m: descriptor for message info
 * @msg: received message header
 *
 * Note: Address is not captured if not requested by receiver.
 */
static void set_orig_addr(struct msghdr *m, struct tipc_msg *msg)
{
	struct sockaddr_tipc *addr = (struct sockaddr_tipc *)m->msg_name;

	if (addr) {
		addr->family = AF_TIPC;
		addr->addrtype = TIPC_ADDR_ID;
		addr->addr.id.ref = msg_origport(msg);
		addr->addr.id.node = msg_orignode(msg);
		addr->addr.name.domain = 0; /* could leave uninitialized */
		addr->scope = 0;        /* could leave uninitialized */
		m->msg_namelen = sizeof (struct sockaddr_tipc);
	}
}


/**
 * put_cmsg - add ancillary data object to control msghdr structure
 */

static unsigned long put_cmsg(void *m, int level, int type, int len, void *data) {
	struct msghdr *msg = (struct msghdr *)m;
	struct cmsghdr *cm = (struct cmsghdr *)msg->msg_control;
	struct cmsghdr cmhdr;
	int cmlen = CMSG_LEN(len);
	int err = 0;

	ASSERT((cm != NULL) && (msg->msg_controllen != 0));

	if (msg->msg_controllen < sizeof (struct cmsghdr)) {
		msg->msg_flags |= MSG_CTRUNC;
		return (0);
	} else if (msg->msg_controllen < cmlen) {
		msg->msg_flags |= MSG_CTRUNC;
		cmlen = msg->msg_controllen;
	}
	cmhdr.cmsg_level = level;
	cmhdr.cmsg_type = type;
	cmhdr.cmsg_len = cmlen;

	/* TODO: equivalent to solaris "ddi_copyout"? */
	memcpy(&cmhdr, cm, sizeof (cmhdr));
	memcpy(data, CMSG_DATA(cm), cmlen - sizeof (struct cmsghdr));
	cmlen = CMSG_SPACE(len);
	
	/* Move to next placeholder */
	msg->msg_control = (char *)msg->msg_control + cmlen;
	msg->msg_controllen -= cmlen;
	return (err);
}




/*
 * anc_data_recv - optionally capture ancillary data for received message
 * @m: descriptor for message info
 * @msg: received message header
 * @tport: TIPC port associated with message
 *
 * Note: Ancillary data is not captured if not requested by receiver.
 *
 * Returns 0 if successful, otherwise errno
 */
static int
anc_data_recv(struct msghdr *m, struct tipc_msg *msg, struct tipc_port *tport)
{
	u32 anc_data[3];
	u32 err;
	u32 dest_type;
	int has_name;
	int res;

	if (likely(m->msg_controllen == 0))
		return (0);

	/* Optionally capture errored message object(s) */

	err = msg ? msg_errcode(msg) : 0;
	if (unlikely(err)) {
		anc_data[0] = err;
		anc_data[1] = msg_data_sz(msg);
		if ((res = put_cmsg(m, SOL_TIPC, TIPC_ERRINFO, 8, anc_data)))
			return (res);
		if (anc_data[1] &&
		    (res = put_cmsg(m, SOL_TIPC, TIPC_RETDATA, anc_data[1],
				    msg_data(msg))))
			return (res);
	}

	/* Optionally capture message destination object */

	dest_type = msg ? msg_type(msg) : TIPC_DIRECT_MSG;
	switch (dest_type) {
	case TIPC_NAMED_MSG:
		has_name = 1;
		anc_data[0] = msg_nametype(msg);
		anc_data[1] = msg_namelower(msg);
		anc_data[2] = msg_namelower(msg);
		break;
	case TIPC_MCAST_MSG:
		has_name = 1;
		anc_data[0] = msg_nametype(msg);
		anc_data[1] = msg_namelower(msg);
		anc_data[2] = msg_nameupper(msg);
		break;
	case TIPC_CONN_MSG:
		has_name = (tport->conn_type != 0);
		anc_data[0] = tport->conn_type;
		anc_data[1] = tport->conn_instance;
		anc_data[2] = tport->conn_instance;
		break;
	default:
		has_name = 0;
	}
	if (has_name &&
	    (res = put_cmsg(m, SOL_TIPC, TIPC_DESTNAME, 12, anc_data)))
		return (res);

	return (0);
}

/*
 * dest_name_check - verify user is permitted to send to specified port name
 * @dest: destination address
 * @m: descriptor for message to be sent
 *
 * Prevents restricted configuration commands from being issued by
 * unauthorized users.
 *
 * Returns 0 if permission is granted, otherwise errno
 */
static int
dest_name_check(struct sockaddr_tipc *dest, struct msghdr *m)
{
//	struct tipc_cfg_msg_hdr hdr;

	if (likely(dest->addr.name.name.type >= TIPC_RESERVED_TYPES))
		return (0);
	if (likely(dest->addr.name.name.type == TIPC_TOP_SRV))
		return (0);
	if (likely(dest->addr.name.name.type != TIPC_CFG_SRV))
		return (-EACCES);

#if 0  // TODO: permission check, permission always granted for now
#ifndef SOLARIS
	if (copy_from_user(&hdr, m->msg_iov[0].iov_base, sizeof (hdr)))
		return (-EFAULT);
	if ((ntohs(hdr.tcm_type) & 0xC000) && (!capable(CAP_NET_ADMIN)))
		return (-EACCES);
#endif
#endif

	return (0);
}


/*
 * advance_rx_queue - discard first buffer in socket receive queue
 *
 * Caller must hold socket lock
 */
static void advance_rx_queue(struct sock *sk)
{
	ASSERT(sk->owner);

	/* using sockbuf queue */
    buf_discard(__skb_dequeue(&sk->sk_socket->so_rcv));
	atomic_dec(&tipc_queue_size);
}


/*
 * discard_rx_queue - discard all buffers in socket receive queue
 *
 * Caller must hold socket lock
 */
static void discard_rx_queue(struct sock *sk)
{
	struct sk_buff *buf;
	ASSERT(sk->owner);

	/*  using sockbuf queue */
    while ((buf = __skb_dequeue(&sk->sk_socket->so_rcv))) {
		atomic_dec(&tipc_queue_size);
		buf_discard(buf);
	}
}


/*
 * reject_rx_queue - reject all buffers in socket receive queue
 *
 * Caller must hold socket lock
 */
static void reject_rx_queue(struct sock *sk)
{
	struct sk_buff *buf;
	ASSERT(sk->owner);

	/* using sockbuf queue */
    while ((buf = __skb_dequeue(&sk->sk_socket->so_rcv))) {
        if(buf->data)
            tipc_reject_msg(buf, TIPC_ERR_NO_PORT);
		atomic_dec(&tipc_queue_size);
	}
}




/**
 * tipc_create - create a TIPC socket
 * @so: pre-allocated socket structure
 * @protocol: protocol indicator (must be 0)
 *
 * This routine creates additional data structures used by the TIPC socket,
 * initializes them, and links them together.
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_create (struct socket *sock, int protocol)
{
	struct sock *sk;
	struct tipc_port *tport;
	uint32_t portref;
	int state = 0;	
	int err;

	PRINT_DBG("%s: protocol %d, type: %d\n", __FUNCTION__, protocol, sock->so_type);

	/* Validate arguments */
	
	if (unlikely(protocol != 0))
		return -EPROTONOSUPPORT;

	switch (sock->so_type) {
	case SOCK_STREAM:
	case SOCK_SEQPACKET:
		state = SS_UNCONNECTED;
		break;
	case SOCK_DGRAM:
	case SOCK_RDM:
		state = SS_READY;
		break;
	default:
		return -EPROTOTYPE;
	}


	/* Allocate socket's protocol area */

	sk = malloc(sizeof(struct sock));
	if (sk == NULL)
	    return -ENOMEM;

	/* Allocate TIPC port for socket to use */
	
	portref = tipc_createport_raw (NULL, &dispatch, &wakeupdispatch,
	                  TIPC_LOW_IMPORTANCE, 0);
	if (unlikely(portref == 0)) {
	    return -ENOMEM;
	}

	/* Finish initializing socket data structures */

	tport = tipc_get_port (portref);
	if (state)	
		sock_state(sk) = state;
	
	sk->sk_rcvtimeo = CONN_TIMEOUT_DEFAULT;
	sk->sk_backlog_rcv = backlog_rcv;
	sk->sk_user_data = (void*)tport;

	if (sock_state(sk) == SS_READY) {
	    tipc_set_portunreturnable(portref, 1);
	    if (sock->so_type == SOCK_DGRAM)
	        tipc_set_portunreliable(portref, 1);
	}

	/* Link PCB to socket & vice versa */
	sock->so_pcb = (void*)sk;
	sk->sk_socket = sock;

	/* Link PCB to TIPC port & vice versa */
	tport->usr_handle = sk;

	/* Initialize other data */
	pthread_mutex_init (&sk->sk_lock, NULL);
	sk->sk_sleep = 0;
	sk->sk_backlog.head = sk->sk_backlog.tail = NULL;
	sk->owner = 0;

	if (err = skb_queue_init (sock, TIPC_TX_BUFSIZE, TIPC_RX_BUFSIZE) != EOK)
		return err;

#if 0
	/* 
	 * "sendto" function should work with both connection and connectionless sockets.
	 * With connectionless socket (SOCK_DGRAM, SOCK_RDM) sendto has a destination addr parameter.
	 * With connection oriented socket (SOCK_SEQPACKET, SOCK_STREAM) sendto's destination addr
	 * parameter can be used to do a "implicit connection", that is, no need to call connect 
	 * explicitly.
	 * 
	 * The problem is, BSD socket layer rejects the sendto if the socket type requires a connection
	 * as determined by the protosw flags PR_CONNREQUIRED. To get around this check, we set the 
	 * flag to SS_ISCONFIRMING so that the check is bypassed. 
	 *
	 *
	 * NOTE: not supporting implicit connections. must call connect if you use connection oriented sockets! 
	 */
	sock->so_state |= SS_ISCONFIRMING;
#endif

	atomic_inc(&tipc_user_count);
	return EOK;
}


/**
 * release - destroy a TIPC socket
 * @sock: socket to destroy
 *
 * This routine cleans up any messages that are still queued on the socket.
 * For DGRAM and RDM socket types, all queued messages are rejected.
 * For SEQPACKET and STREAM socket types, the first message is rejected
 * and any others are discarded.  (If the first message on a STREAM socket
 * is partially-read, it is discarded and the next one is rejected instead.)
 *
 * NOTE: Rejected messages are not necessarily returned to the sender!  They
 * are returned or discarded according to the "destination droppable" setting
 * specified for the message by the sender.
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_release (struct socket *sock)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport;
	struct sk_buff *buf;
	int res;

	/*
	 * Exit if socket isn't fully initialized (occurs when a failed accept()
	 * releases a pre-allocated child socket that was never used)
	 */

	//PRINT_DBG("%s\n", __FUNCTION__);
	if (sk == NULL)
	{
		if (sock)
		{
			res = 0;
			goto SOCK_RELEASE_EXIT;	
		}		
		return 0;
	}

	tport = sock_port(sk);
	lock_sock(sk);

	/*
	 * Reject all unreceived messages, except on an active connection
	 * (which disconnects locally & sends a 'FIN+' to peer)
	 */

	PRINT_DBG("%s: state: %d\n", __FUNCTION__, sock_state(sk));
	while (sock_state(sk) != SS_DISCONNECTING) {
		buf = __skb_dequeue(&sock->so_rcv);
		if (buf == NULL)
			break;
		atomic_dec(&tipc_queue_size);
		if (!buf->data || TIPC_SKB_CB(buf)->handle != msg_data(buf_msg(buf)))
			buf_discard(buf);
		else {
			if ((sock_state(sk) == SS_CONNECTING) ||
				(sock_state(sk) == SS_CONNECTED)) {
				sock_state(sk) = SS_DISCONNECTING;
				tipc_disconnect(tport->ref);
			}
			tipc_reject_msg(buf, TIPC_ERR_NO_PORT);
		}
	}

	/*
	 * Delete TIPC port; this ensures no more messages are queued
	 * (also disconnects an active connection & sends a 'FIN-' to peer)
	 */

	PRINT_DBG("%s: tipc_deleteport\n", __FUNCTION__);
	res = tipc_deleteport(tport->ref);

	/* Discard any remaining (connection-based) messages in receive queue */

	discard_rx_queue(sk);

	/* Reject any messages that accumulated in backlog queue */

	sock_state(sk) = SS_DISCONNECTING;
	release_sock(sk);
	ASSERT(!sock_mutex_owned(sk));

    atomic_dec(&tipc_user_count);
	PRINT_DBG("%s: exit\n", __FUNCTION__);

	/* free the socket */
	pthread_mutex_destroy (&sk->sk_lock);
	memset(sk, 0, sizeof(struct sock));
	free(sk);

SOCK_RELEASE_EXIT:
	sbrelease (&sock->so_rcv, sock);
	sbrelease (&sock->so_snd, sock);
	sofree(sock);
	return res;
}



/**
 * bind - associate or disassocate TIPC name(s) with a socket
 * @sock: socket structure
 * @mbuf: mbuf with socket address describing name(s) and desired operation
 *
 * Name and name sequence binding is indicated using a positive scope value;
 * a negative scope value unbinds the specified name.  Specifying no name
 * (i.e. a socket address length of 0) unbinds all names from the socket.
 *
 * Returns 0 on success, errno otherwise
 *
 * NOTE: This routine doesn't need to take the socket lock since it doesn't
 *       access any non-constant socket information.
 */

int tipc_sock_bind (struct socket *sock, struct mbuf *nam)
{
	struct sockaddr_tipc *addr = mtod(nam, struct sockaddr_tipc *);
	int uaddr_len = nam->m_len;

	struct sock *sk = (struct sock*)sock->so_pcb;
	u32 portref = sock_port(sk)->ref;
	PRINT_DBG("%s: len: %d, fam: %d, scope: %d\n", 
		__FUNCTION__, uaddr_len, addr->family, addr->scope);

	if (unlikely(!uaddr_len))
		return tipc_withdraw(portref, 0, NULL);

	if (uaddr_len < sizeof(struct sockaddr_tipc))
		return -EINVAL;

	/* MLL@SRD TODO: problem - sys_bind function in the 
	 *               io-pkt stack modifies mbuf data so 
	 *               that the following compare fails. To 
	 *               be specific, sys_bind thinks that the
	 *               data is of type "struct sockaddr" so it
	 *               adds data length into the buf where the 
	 *               tipc family data is located!
	 */
	if (addr->family != AF_TIPC)
		return -EAFNOSUPPORT;

	PRINT_DBG("%s: addrtype: %d\n", __FUNCTION__, addr->addrtype);
	if (addr->addrtype == TIPC_ADDR_NAME)
		addr->addr.nameseq.upper = addr->addr.nameseq.lower;
	else if (addr->addrtype != TIPC_ADDR_NAMESEQ)
		return -EAFNOSUPPORT;

	PRINT_DBG("%s: scope: %d\n", __FUNCTION__, addr->scope);
	return (addr->scope > 0) ?
		tipc_publish(portref, addr->scope, &addr->addr.nameseq) :
		tipc_withdraw(portref, -addr->scope, &addr->addr.nameseq);
}


/**
 * listen - allow socket to listen for incoming connections
 * @sock: socket structure
 * @len: (unused)
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_listen (struct socket *sock)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	int res;

	lock_sock(sk);

	if (sock_state(sk) == SS_READY)
		res = -EOPNOTSUPP;
	else if (sock_state(sk) != SS_UNCONNECTED)
		res = -EINVAL;
	else {
		sock_state(sk) = SS_LISTENING;
		res = 0;
	}

	release_sock(sk);
	ASSERT(!sock_mutex_owned(sk));
	return res;
}



/**
 * accept - wait for connection request
 * @sock: listening socket
 * @newsock: new socket that is to be connected
 * @flags: file-related flags associated with socket
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_accept (struct socket *new_sock, struct sk_buff *buf, struct mbuf *mb)
{
	struct sock *new_sk = (struct sock*)new_sock->so_pcb;
	struct tipc_port *new_tport = sock_port(new_sk);
	u32 new_ref = new_tport->ref;
	struct tipc_msg *msg = buf_msg(buf);
	struct tipc_portid id;
	int res = EOK;

	PRINT_DBG("%s: start sock 0x%x\n", __FUNCTION__, (int)new_sock);
	lock_sock(new_sk);

	PRINT_DBG("%s: connect new socket to it's peer\n", __FUNCTION__);
	/* Connect new socket to it's peer */
		   
	id.ref = msg_origport(msg);
	id.node = msg_orignode(msg);
	tipc_connect2port(new_ref, &id);
	sock_state((struct sock *)new_sock->so_pcb) = SS_CONNECTED;

	PRINT_DBG("%s: tipc_set_portimportance\n", __FUNCTION__);
	tipc_set_portimportance(new_ref, msg_importance(msg));
	if (msg_named(msg)) {
		new_tport->conn_type = msg_nametype(msg);
		new_tport->conn_instance = msg_nameinst(msg);
	}

	/*
	 * Respond to 'SYN-' by discarding it & returning 'ACK'-.
	 * Respond to 'SYN+' by queuing it on new socket.
	 */

	msg_dbg(msg, "<ACC<: ");
	PRINT_DBG("%s: <ACC<:\n", __FUNCTION__);
	if (!msg_data_sz(msg)) {
		struct msghdr mh;
		memset(&mh, 0, sizeof (mh));

		PRINT_DBG("%s: tipc_sock_send_packet\n", __FUNCTION__);
		tipc_sock_send_packet(new_sock, &mh, 0, NULL);

		MBUF_FREE(mb);
	} else {
		PRINT_DBG("%s: sbappend to sock 0x%x\n", __FUNCTION__, (int)new_sock);
		sbappend(&new_sock->so_rcv, mb);	
	}

	release_sock(new_sk);
	ASSERT(!sock_mutex_owned(new_sk));
	PRINT_DBG("%s: exit %d\n", __FUNCTION__, res);
	return (res);
}



/**
 * connect - establish a connection to another TIPC port
 * @sock: socket structure
 * @dest: socket address for destination port
 * @destlen: size of socket address data structure
 * @flags: file-related flags associated with socket
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_connect (struct socket *sock, struct mbuf *nam, int flags)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct sockaddr_tipc *dst = mtod(nam, struct sockaddr_tipc *);
	struct msghdr m;
	struct tipc_msg *msg;
	struct mbuf *buf;
	int destlen = nam->m_len;
	int res;
	
	PRINT_DBG("%s: state %d\n", __FUNCTION__, sock_state(sk));
	lock_sock(sk);
	
	/* For now, TIPC does not allow use of connect() with DGRAM/RDM types */
	
	if (sock_state(sk) == SS_READY) {
	    res = EOPNOTSUPP;
	    goto exit;
	}
	PRINT_DBG("%s: state is ready\n", __FUNCTION__);

	/* For now, TIPC does not support the non-blocking form of connect() */

	if (flags & O_NONBLOCK) {
	    res = EWOULDBLOCK;
	    goto exit;
	}
	PRINT_DBG("%s: state is blocking\n", __FUNCTION__);

	/* Issue Posix-compliant error code if socket is in the wrong state */

	if (sock_state(sk) == SS_LISTENING) {
	    res = EOPNOTSUPP;
	    goto exit;
	}
	if (sock_state(sk) == SS_CONNECTING) {
	    res = EALREADY;
	    goto exit;
	}
	if (sock_state(sk) != SS_UNCONNECTED) {
	    res = EISCONN;
	    goto exit;
	}


	/*
	 * Reject connection attempt using multicast address
	 *
	 * Note: send_msg() validates the rest of the address fields,
	 *       so there's no need to do it here
	 */
		 
	if (dst->addrtype == TIPC_ADDR_MCAST) {
		res = EINVAL;
		goto exit;
	}
	
	/* let BSD stack know that we are attempting to connect 
	 * sonewconn called
	 */
	soisconnecting(sock);

	/* Reject any messages already in receive queue (very unlikely) */

	reject_rx_queue(sk);

	/* Send a 'SYN-' to destination */

	PRINT_DBG("%s: send_msg (%d bytes)\n", __FUNCTION__, sizeof(m));
	memset(&m, 0, sizeof(m));
	m.msg_name = dst;
	m.msg_namelen = destlen;

	if ((res = tipc_sock_send_msg(sock, &m, 0, NULL)) < 0) {
		goto exit;
	}

	/* Wait until an 'ACK' or 'RST' arrives, or a timeout occurs */

	PRINT_DBG("%s: is queue empty? %d\n", __FUNCTION__, skb_queue_empty(&sock->so_rcv));
	while (skb_queue_empty(&sock->so_rcv) && (sock_state(sk) == SS_CONNECTING)) 
	{
		sock_mutex_enter(sk);
		handle_backlog(sk);
		if (skb_queue_empty(&sock->so_rcv) && (sock_state(sk) == SS_CONNECTING))
		{
			sk->owner = NULL;
			if (sk->waiters)
				sock_wakeup(sk);

			PRINT_DBG("%s: sock_wait enter (sock 0x%x), timeout in %d ms\n", __FUNCTION__, (int)sock, sk->sk_rcvtimeo);
			res = sock_wait (sk, sk->sk_rcvtimeo);
			PRINT_DBG("%s: sock_wait exit %d (sock 0x%x)\n", __FUNCTION__, res, (int)sock);
			if (res == EOK) {
				lock_sock(sk);
			} else if (res == EAGAIN) {
				warn("%s: sock_wait error, %s\n", __FUNCTION__, strerror(res));
				sock_mutex_exit(sk);
				res = ETIMEDOUT;
				goto exit_immediately;
			} else {
				warn("%s: sock_wait error, %s\n", __FUNCTION__, strerror(res));
				sock_mutex_exit(sk);
				res = EINTR;
				goto exit_immediately;
			}
		}
		sock_mutex_exit(sk);
	}

	buf = skb_peek(&sock->so_rcv);
	if (buf != NULL) {
		msg = mtod(buf, struct tipc_msg *);
		res = auto_connect(sock, msg);
		PRINT_DBG("%s: auto_connect %d\n", __FUNCTION__, res);
		if (!res) {
			if (!msg_data_sz(msg))
				advance_rx_queue(sk);
		}
		else {
			PRINT_DBG("%s: auto_connect failed!!!\n", __FUNCTION__);
		}
	} else { 
		if (sock_state(sk) == SS_CONNECTED) {
			res = EISCONN;
		} else {
			res = ECONNREFUSED;
		}
	}

exit:
	release_sock(sk);
exit_immediately:
	ASSERT(!sock_mutex_owned(sk));
	PRINT_DBG("%s: exit %d\n", __FUNCTION__, res);
	return res;
}



/**
 * shutdown - shutdown socket connection
 * @sock: socket structure
 * @how: direction to close (must be SHUT_RDWR)
 *
 * Terminates connection (if necessary), then purges socket's receive queue.
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_shutdown (struct socket *sock)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	struct sk_buff *buf;
	int res = EOK;
	int do_pollwakeup = 0;

	PRINT_DBG("%s: start\n", __FUNCTION__);
	lock_sock(sk);

	PRINT_DBG("%s: state: %d\n", __FUNCTION__, sock_state(sk));
	switch (sock_state(sk)) {
	case SS_CONNECTING:
	case SS_CONNECTED:

		PRINT_DBG("%s: state is connecting/connected\n", __FUNCTION__);
		/* Disconnect and send a 'FIN+' or 'FIN-' message to peer */
restart:	
		if ((buf = __skb_dequeue(&sock->so_rcv))) {
			atomic_dec(&tipc_queue_size);
			if (!buf->data || 
                TIPC_SKB_CB(buf)->handle !=
				msg_data(buf_msg(buf))) {
				buf_discard(buf);
				goto restart;
			}
			PRINT_DBG("%s: disconnect and reject msg\n", __FUNCTION__);
			tipc_disconnect(tport->ref);
			tipc_reject_msg(buf, TIPC_CONN_SHUTDOWN);
		} else {
			PRINT_DBG("%s: tipc_shutdown\n", __FUNCTION__);
			tipc_shutdown(tport->ref);
		}

		sock_state(sk) = SS_DISCONNECTING;

	/* fall through */

	case SS_DISCONNECTING:
	
		/* Discard any unreceived messages; wake up sleeping tasks */

		PRINT_DBG("%s: state is disconecting\n", __FUNCTION__);
		discard_rx_queue(sk);
		sock_wakeup(sk);
		do_pollwakeup = 1;
		res = 0;
		soisdisconnected(sock);
		break;

	default:
		PRINT_DBG("%s: state is ENOTCONN\n", __FUNCTION__);
		res = -ENOTCONN;
	}

	release_sock(sk);
	ASSERT(!sock_mutex_owned(sk));

	if (do_pollwakeup)
	{
		PRINT_DBG("%s: wakeup 0x%x\n", __FUNCTION__, (int)sock);
		sorwakeup(sock);
		wakeup((caddr_t)&sock->so_timeo);
	}
	PRINT_DBG("%s: exit\n", __FUNCTION__);
	return (res);
}




/*
 * send_msg - send message in connectionless manner
 * @iocb: unused (if NULL, indicates that socket lock is already held)
 * @sock: socket structure
 * @m: message to send
 * @total_len: length of message
 *
 * Message must have an destination specified explicitly.
 * Used for SOCK_RDM and SOCK_DGRAM messages,
 * and for 'SYN' messages on SOCK_SEQPACKET and SOCK_STREAM connections.
 * (Note: 'SYN+' is prohibited on SOCK_STREAM.)
 *
 * Returns the number of bytes sent on success, or errno otherwise
 */

int tipc_sock_send_msg (struct socket *sock, struct msghdr *m, int total_len, void *iocb)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	struct sockaddr_tipc *dest = (struct sockaddr_tipc *)m->msg_name;
	int needs_conn;
	int res = -EINVAL;

	PRINT_DBG("%s: family: %d, addrtype: %d, type: %d, instance: %d, domain: %d\n",
		__FUNCTION__, dest->family, dest->addrtype, dest->addr.name.name.type, 
		dest->addr.name.name.instance, dest->addr.name.domain);

	PRINT_DBG("%s: total_len: %d bytes\n", __FUNCTION__, total_len);	
	if (unlikely(!dest))
	    return (-EDESTADDRREQ);
	if (dest->family != AF_TIPC)
	    return (-EINVAL);

	PRINT_DBG("%s: lock_sock\n", __FUNCTION__);	
	if (iocb)
	    lock_sock(sk);

	needs_conn = (sock_state(sk) != SS_READY);
	PRINT_DBG("%s: needs_conn: %d, sock_state %d\n", __FUNCTION__, needs_conn, sock_state(sk));	
	if (unlikely(needs_conn)) {
		if (sock_state(sk) == SS_LISTENING) {
			res = -EPIPE;
			goto exit;
		}
		if (sock_state(sk) != SS_UNCONNECTED) {
			res = -EISCONN;
			goto exit;
		}
		if ((tport->published) ||
			((sock->so_type == SOCK_STREAM) && (total_len != 0))) {
			res = -EOPNOTSUPP;
			goto exit;
		}
		if (dest->addrtype == TIPC_ADDR_NAME) {
			tport->conn_type = dest->addr.name.name.type;
			tport->conn_instance = dest->addr.name.name.instance;
		}

		/* Abort any pending connection attempts (very unlikely) */

		reject_rx_queue(sk);
	}

	do {
		if (dest->addrtype == TIPC_ADDR_NAME) {
			PRINT_DBG("%s: addrtype TIPC_ADDR_NAME\n", __FUNCTION__);	
		    if ((res = dest_name_check(dest, m)))
		        break;
		    res = tipc_send2name(tport->ref,
		        &dest->addr.name.name, dest->addr.name.domain,
		        m->msg_iovlen, m->msg_iov);
			PRINT_DBG("%s: addrtype TIPC_ADDR_NAME ret: %d\n", __FUNCTION__, res);	
		} else if (dest->addrtype == TIPC_ADDR_ID) {
			PRINT_DBG("%s: addrtype TIPC_ADDR_ID\n", __FUNCTION__);	
		    res = tipc_send2port(tport->ref, &dest->addr.id,
		        m->msg_iovlen, m->msg_iov);
		} else if (dest->addrtype == TIPC_ADDR_MCAST) {
			PRINT_DBG("%s: addrtype TIPC_ADDR_MCAST\n", __FUNCTION__);	
		    if (needs_conn) {
		        res = -EOPNOTSUPP;
		        break;
		    }
		    if ((res = dest_name_check(dest, m)))
		        break;
		    res = tipc_multicast(tport->ref,
		        &dest->addr.nameseq, 0, m->msg_iovlen,
		        m->msg_iov);
		}
		PRINT_DBG("%s: res: %d\n", __FUNCTION__, res);	
		if (likely(res != -ELINKCONG)) {
		    if (needs_conn && (res >= 0)) {
		        sock_state(sk) = SS_CONNECTING;
		    }
		    break;
		}
		if (m->msg_flags & MSG_DONTWAIT) {
		    res = -EWOULDBLOCK;
		    break;
		}
		
		while (tport->congested) {
			PRINT_DBG("%s: congested? %d\n", __FUNCTION__, tport->congested);	
			sock_mutex_enter(sk);
			handle_backlog(sk);
			if (tport->congested) {
				sk->owner = NULL;
				if (sk->waiters)
					sock_wakeup(sk);
				PRINT_DBG("%s: sock_wait\n", __FUNCTION__);	
				res = sock_wait(sk, 1000);
				PRINT_DBG("%s: sock_wait res: %d\n", __FUNCTION__, res);	
				ASSERT(res >= 0);
				if ((res == 0) || (res == EAGAIN))
					lock_sock(sk);
				else {
					PRINT_DBG("%s: sock_wait error: %d\n", __FUNCTION__, res);	
					sock_mutex_exit(sk);
					res = -EINTR;
					if (!iocb) {
						/*
						 * we didn't do the lock_sock(),
						 * so restore it before leaving
						 */
						lock_sock(sk);
					}
					goto exit_immediately;
				}
			}
			sock_mutex_exit(sk);
		}
	} while(1);

exit:
	if (iocb)
		release_sock(sk);
exit_immediately:
	ASSERT(!sock_mutex_owned(sk));
	PRINT_DBG("%s: exit\n", __FUNCTION__);	
	return (res);
}

/*
 * send_packet - send a connection-oriented message
 * @iocb: unused (if NULL, indicates that socket lock is already held)
 * @sock: socket structure
 * @m: message to send
 * @total_len: length of message
 *
 * Used for SOCK_SEQPACKET messages and SOCK_STREAM data.
 *
 * Returns the number of bytes sent on success, or errno otherwise
 */

int tipc_sock_send_packet (struct socket *sock, struct msghdr *m, int total_len, void *iocb)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	struct sockaddr_tipc *dest = (struct sockaddr_tipc *)m->msg_name;
	int res = -EINVAL;

	PRINT_DBG("%s: start\n", __FUNCTION__);

	/* Handle implied connection establishment */

	if (unlikely(dest))
	    return (tipc_sock_send_msg(sock, m, total_len, iocb));

	if (iocb)
	    lock_sock(sk);

	do {
		if (unlikely(sock_state(sk) != SS_CONNECTED)) {
			if (sock_state(sk) == SS_DISCONNECTING)
				res = -EPIPE;
			else
				res = -ENOTCONN;
			break;
		}

		PRINT_DBG("%s: tipc_send %d bytes, %d iovs -------------\n", __FUNCTION__, total_len, m->msg_iovlen);	
		res = tipc_send(tport->ref, m->msg_iovlen, m->msg_iov);
		if (likely(res != -ELINKCONG)) {
		    break;
		}
		if (m->msg_flags & MSG_DONTWAIT) {
		    res = -EWOULDBLOCK;
		    break;
		}
		
		while (tport->congested && tport->connected) {
			PRINT_DBG("%s: congested? %d, connected? %d\n", __FUNCTION__, tport->congested, tport->connected);	
			sock_mutex_enter(sk);
		    handle_backlog(sk);
		    if (tport->congested && tport->connected) {
		        sk->owner = NULL;
				if (sk->waiters)
					sock_wakeup(sk);
				PRINT_DBG("%s: sock_wait\n", __FUNCTION__);	
				res = sock_wait(sk, 1000);
				PRINT_DBG("%s: sock_wait res: %d\n", __FUNCTION__, res);	
				if ((res == 0) || (res == EAGAIN))
					lock_sock(sk);
				else {
					PRINT_DBG("%s: sock_wait error: %d\n", __FUNCTION__, res);	
					sock_mutex_exit(sk);
					res = -EINTR;
					if (!iocb) {
						/*
						 * we didn't do the lock_sock(),
						 * so restore it before leaving
						 */
						PRINT_DBG("%s: sock_wait res %d, lock_sock\n");
						lock_sock(sk);
					}
				    goto exit_immediately;
				}
			}
			sock_mutex_exit(sk);
		}
    } while (1);

	if (iocb)
		release_sock(sk);
exit_immediately:
	PRINT_DBG("%s: exit\n", __FUNCTION__);
	ASSERT(!sock_mutex_owned(sk));
	return (res);
}
  

/*
 * recv_msg - receive packet-oriented message
 * @m: descriptor for message info
 * @buf_len: total size of user buffer area
 * @flags: receive flags
 *
 * Used for SOCK_DGRAM, SOCK_RDM, and SOCK_SEQPACKET messages.
 * If the complete message doesn't fit in user area, truncate it.
 *
 * Returns size of returned message data, errno otherwise
 */

int tipc_sock_recv_msg (struct socket *sock)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);

	PRINT_DBG("tipc_sock_recv_msg: sock 0x%x\n", (int)sock);

//	lock_sock(sk);

	/* Consume received message (optional) */

	if ((sock_state(sk) != SS_READY) &&
		(++tport->conn_unacked >= TIPC_FLOW_CONTROL_WIN)) {

		tipc_acknowledge(tport->ref, tport->conn_unacked);
	}

	/* TODO: do we need tipc_queue_size to keep track of queue_size? can we just use "skb_queue_len"?? */
	atomic_dec(&tipc_queue_size);

//    release_sock(sk);
//    ASSERT(!sock_mutex_owned(sk));
	PRINT_DBG("tipc_sock_recv_msg: exit\n", (int)sock);
    return EOK;
}


/**
 * get_name - get port ID of socket or peer socket
 * @sock: socket structure
 * @uaddr: area for returned socket address
 * @uaddr_len: area for returned length of socket address
 * @peer: 0 to obtain socket name, 1 to obtain peer socket name
 *
 * Returns 0 on success, errno otherwise
 *
 * NOTE: This routine doesn't need to take the socket lock since it doesn't
 *       access any non-constant socket information.
 */

static int tipc_sock_get_name(struct socket *sock, struct sockaddr_tipc *addr,
				    		  int *uaddr_len, int peer)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	u32 portref = sock_port(sk)->ref;
	u32 res;

	if (peer == SET_PEER_ADDR) {
	    if ((res = tipc_peer(portref, &addr->addr.id))) {
	        return res;
		}	
	} else {
	    tipc_ownidentity(portref, &addr->addr.id);
	}

	*uaddr_len = sizeof(*addr);
	addr->addrtype = TIPC_ADDR_ID;
	addr->family = AF_TIPC;
	addr->scope = 0;
	addr->addr.name.domain = 0;

	return EOK;
}



/**
 * setsockopt - set socket option
 * @sock: socket structure
 * @lvl: option level
 * @opt: option identifier
 * @ov: pointer to new option value
 * @ol: length of option value
 *
 * For stream sockets only, accepts and ignores all IPPROTO_TCP options
 * (to ease compatibility).
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_setsockopt (struct socket *sock, int lvl, int opt, char *ov, int ol) 
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	u32 value;
	int res = EOK;
	
	//PRINT_DBG("%s: lvl: %d, opt: %d\n", __FUNCTION__, lvl, opt);	
	if ((lvl == IPPROTO_TCP) && (sock->so_type == SOCK_STREAM))
		return (0);
	if (lvl != SOL_TIPC)
		return (-ENOPROTOOPT);
	if (ol < sizeof (value))
		return (-EINVAL);
	
	value = *((u32 *)ov);

	lock_sock(sk);

	switch (opt) {
	case TIPC_IMPORTANCE:
		res = tipc_set_portimportance(tport->ref, value);
		break;
	case TIPC_SRC_DROPPABLE:
		if (sock->so_type != SOCK_STREAM)
			res = tipc_set_portunreliable(tport->ref, value);
		else
			res = -ENOPROTOOPT;
		break;
	case TIPC_DEST_DROPPABLE:
		res = tipc_set_portunreturnable(tport->ref, value);
		break;
	case TIPC_CONN_TIMEOUT:
		/* value is in milliseconds */
		sk->sk_rcvtimeo = value;
		/* no need to set "res", since already 0 at this point */
		break;
	default:
		res = -EINVAL;
	}

	release_sock(sk);
	ASSERT(!sock_mutex_owned(sk));

	return (res);
}

/**
 * getsockopt - get socket option
 * @sock: socket structure
 * @lvl: option level
 * @opt: option identifier
 * @ov: receptacle for option value
 * @ol: receptacle for length of option value
 *
 * For stream sockets only, returns 0 length result for all IPPROTO_TCP options
 * (to ease compatibility).
 *
 * Returns 0 on success, errno otherwise
 */

int tipc_sock_getsockopt (struct socket *sock, int lvl, int opt, char *ov, int *ol) 
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct tipc_port *tport = sock_port(sk);
	u32 value;
	int res = EOK;

	//PRINT_DBG("%s: lvl: %d, opt: %d\n", __FUNCTION__, lvl, opt);	
	if ((lvl == IPPROTO_TCP) && (sock->so_type == SOCK_STREAM)) {
	   	*ol = 0; 
		return EOK;
	}	
	if (lvl != SOL_TIPC)
	    return (-ENOPROTOOPT);

	lock_sock(sk);


	switch (opt) {
	case TIPC_IMPORTANCE:
		res = tipc_portimportance(tport->ref, &value);
		break;
	case TIPC_SRC_DROPPABLE:
		res = tipc_portunreliable(tport->ref, &value);
		break;
	case TIPC_DEST_DROPPABLE:
		res = tipc_portunreturnable(tport->ref, &value);
		break;
	case TIPC_CONN_TIMEOUT:
		/* value is in milliseconds */
		value = sk->sk_rcvtimeo;
		/* no need to set "res", since already 0 at this point */
		break;
	default:
		res = -EINVAL;
	}

	release_sock(sk);
	ASSERT(!sock_mutex_owned(sk));

	if (res) {
	    /* "get" failed */
		err("%s: failed, %s\n", __FUNCTION__, strerror(res));	
	} else {
		*(unsigned int *)ov = value;
		*ol = sizeof(value);
	}

	return (res);
}


/* setup mhdr struct to point to mbuf data */
static void tipc_mbuf2msghdr (struct socket *sock, struct mbuf *m, struct mbuf *nam, struct msghdr *mh)
{
	int i, total_len;
	struct mbuf *tmpbuf;
	struct iovec *curr_iov;

	mh->msg_name = nam? mtod(nam, void*) : NULL;
	mh->msg_namelen = nam? sizeof(struct sockaddr_tipc) : 0;
	mh->msg_flags = (sock->so_state & SS_NBIO)? MSG_DONTWAIT : 0;

	tmpbuf = m;	
	total_len = 0;
	curr_iov = mh->msg_iov;
	for (i=0; tmpbuf; i++)
	{
		curr_iov->iov_base = mtod(tmpbuf, void*);
		curr_iov->iov_len  = tmpbuf->m_len;
		total_len += curr_iov->iov_len;
		curr_iov++;
		
		tmpbuf = tmpbuf->m_next;
	}
	mh->msg_iovlen = i;

	PRINT_DBG("%s: total_len: %d bytes, num_iovs: %d\n", __FUNCTION__, total_len, mh->msg_iovlen);
	ASSERT(total_len == m->m_pkthdr.len);
}

static int tipc_usrreq(struct socket *sock, int req, struct mbuf *mb, struct mbuf *nam,
						struct mbuf *control, struct lwp *l)
{
	struct sock *sk = (struct sock*)sock->so_pcb;
	struct proc *p;
	int error = EOK;

	/*
	 * see protosw.h:
	 * The arguments to usrreq are:
	 *  (*protosw[].pr_usrreq)(up, req, m, nam, opt, p);
	 * where up is a (struct socket *), req is one of these requests,
	 * m is a optional mbuf chain containing a message,
	 * nam is an optional mbuf chain containing an address,
	 * opt is an optional mbuf containing socket options,
	 * and p is a pointer to the process requesting the action (if any).
	 * The protocol is responsible for disposal of the mbuf chains m and opt,
	 * the caller is responsible for any space held by nam.
	 * A non-zero return from usrreq gives an
	 * UNIX error number which should be passed to higher level software.
	 *
	 */
	
	PRINT_DBG("%s: usrreq: %s\n", __FUNCTION__, prurequests[req]);
	
	if (req == PRU_CONTROL)
		return (EOPNOTSUPP);

	p = l ? LWP_TO_PR(l) : NULL;

	if (sk == 0 && req != PRU_ATTACH && req != PRU_SENSE) {
		error = EINVAL;
		goto release;
	}

	switch (req) {

	case PRU_ATTACH:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		if (sk != 0) {
			error = EISCONN;
			break;
		}
		error = tipc_sock_create(sock, (int)nam);
		break;

	case PRU_DETACH:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		tipc_sock_release(sock);
		break;

	case PRU_BIND:
		ASSERT(l != NULL);
		error = tipc_sock_bind(sock, nam);
		break;

	case PRU_LISTEN:
		error = tipc_sock_listen(sock);
		break;

	case PRU_ACCEPT:
	{
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		soisconnected(sock);
		break;
	}

	case PRU_CONNECT:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		
		error = tipc_sock_connect(sock, nam,
			(sock->so_state & SS_NBIO)? O_NONBLOCK : 0);
		break;

	case PRU_DISCONNECT:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		
		soisdisconnected(sock);
		break;

	case PRU_SHUTDOWN:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		
		socantsendmore(sock);
		error = tipc_sock_shutdown(sock);
		break;

	case PRU_RCVD:

		switch (sock->so_type) {
		case SOCK_DGRAM:
		case SOCK_RDM:
		case SOCK_SEQPACKET:
			error = tipc_sock_recv_msg (sock);
			break;
	
		/* TODO: support for SOCK_STREAM */	
		case SOCK_STREAM:
			error = ENOTSUP;
			break;
		default:
			break;	
		}
		break;

	case PRU_SEND:
	
		PRINT_DBG("%s: PRU_SEND type %d\n", __FUNCTION__, sock->so_type);
		switch (sock->so_type) {
		case SOCK_DGRAM:
		case SOCK_RDM:
		case SOCK_SEQPACKET:
		{
			/* setup our msghdr struct */
			struct msghdr mh;
			struct iovec iov[100];
			memset (&mh, 0, sizeof(struct msghdr));	
			mh.msg_iov = &iov;
			mh.msg_iovlen = 1;

			PRINT_DBG("%s: control 0x%x\n", __FUNCTION__, (int)control);
			if (control && control->m_len) {
				MBUF_FREE(control);
				MBUF_FREE(mb);
				error = EINVAL;
				break;
			}

			/* convert mbuf to msghdr */	
			tipc_mbuf2msghdr (sock, mb, nam, &mh);
			if (sock->so_type == SOCK_SEQPACKET)
				error = tipc_sock_send_packet(sock, &mh, mb->m_pkthdr.len, (void*)1);
			else	
				error = tipc_sock_send_msg(sock, &mh, mb->m_pkthdr.len, (void*)1);
			if (error < 0) {
				error = -error;
				MBUF_FREE(mb);
			}
			else if (error > 0) {
				error = EOK;
			}
			MBUF_FREE(mb);
			break;
		}

		/* TODO: support for SOCK_STREAM */	
		case SOCK_STREAM:
			error = ENOTSUP;
			break;
		default:
			break;
		}
		break;

	case PRU_RCVOOB:
		error = EOPNOTSUPP;
		break;

	case PRU_SENDOOB:
		MBUF_FREE(control);
		MBUF_FREE(mb);
		error = EOPNOTSUPP;
		break;

	case PRU_SOCKADDR:
		error = tipc_sock_get_name (sock, mtod(nam, struct sockaddr_tipc*), 
				    				&nam->m_len, SET_SOCK_ADDR);
		break;

	case PRU_PEERADDR:
		error = tipc_sock_get_name (sock, mtod(nam, struct sockaddr_tipc*),
				    				&nam->m_len, SET_PEER_ADDR);
		break;
        
    case PRU_SENSE:
        if(nam)
        {
            err("Got sense request for TIPC with name\n");
        }
        else
            err("Got sense request for TIPC\n");
        return EOPNOTSUPP;

	default:
		if (control)
			MBUF_FREE(control);	
		if (mb)
			MBUF_FREE(mb);	
		err("User Request ID %d unknown\n", req);
		return EINVAL;	
	}


release:
	return (error);
}



/*
 * tipc_ctloutput - socket options processing
 *
 * @cmd: is one of the actions listed below, so is a (struct socket *),
 * @level: is an indication of which protocol layer the option is intended.
 * @optname: is a protocol dependent socket option request,
 * @optva:l is a pointer to a mbuf-chain pointer, for value-return results.
 * 
 * The protocol is responsible for disposal of the mbuf chain *optval
 * if supplied, the caller is responsible for any space held by *optval, when returned.
 * 
 * Returns 0 on success, errno otherwise 
 */

static int tipc_ctloutput(int cmd, struct socket *sock, int level, int optname, struct mbuf **m)
{
	int error = EOK;

	if (level != SOL_TIPC) {
	    error = ENOPROTOOPT;
	    if ((cmd == PRCO_SETOPT) && (*m != NULL)) {
	    	MBUF_FREE(*m);
		}
	}
	else {
		switch (cmd) {
		case PRCO_SETOPT:
			if (*m != NULL) {	
				error = tipc_sock_setsockopt (sock, level, optname, 
							(mtod(*m, char*)), (*m)->m_len);
			}
			else {
				err("%s: mbuf is null!\n", __FUNCTION__);
				error = EINVAL;	
				MBUF_FREE(*m);
			}
			break;

		case PRCO_GETOPT:
			/* need to allocate mbuf for opt values */	
			*m = (struct mbuf*) m_get(M_WAIT, MT_SOOPTS);

			error = tipc_sock_getsockopt (sock, level, optname, 
							(mtod(*m, char*)), &((*m)->m_len));
			break;
		default:
			PRINT_DBG("%s: unknown cmd: %d\n", __FUNCTION__, cmd);
			break;	
		}
	}
	return (error);
}



static void tipc_init_domain (void)
{
}

static void tipc_init_dgram (void)
{
}

static void tipc_init_rdm (void)
{
}

static void tipc_init_seqpacket (void)
{
}

static void tipc_init_stream (void)
{
}


/* forward define */
struct domain tipcdomain;

const struct protosw tipcsw[] = {
{
	SOCK_DGRAM, &tipcdomain, 0, PR_ATOMIC|PR_ADDR,
	0, 0, 0, &tipc_ctloutput,
	&tipc_usrreq,
	&tipc_init_dgram, 0, 0, 0, 
},
{
	SOCK_RDM, &tipcdomain, 0,	PR_ATOMIC|PR_ADDR,
	0, 0, 0, &tipc_ctloutput,
	&tipc_usrreq,
	&tipc_init_rdm, 0, 0, 0,
},
{
	SOCK_SEQPACKET, &tipcdomain, 0,	PR_CONNREQUIRED|PR_WANTRCVD|PR_LISTEN|PR_ATOMIC|PR_ABRTACPTDIS,
	0, 0, 0, &tipc_ctloutput,
	&tipc_usrreq,
	&tipc_init_seqpacket, 0, 0, 0,
},
#if 0 // TODO: support for SOCK_STREAM
{
	SOCK_STREAM, &tipcdomain, 0, PR_CONNREQUIRED|PR_WANTRCVD|PR_LISTEN|PR_ABRTACPTDIS,
	0, 0, 0, &tipc_ctloutput,
	&tipc_usrreq,
	&tipc_init_stream, 0, 0, 0,
},
#endif
};

struct domain tipcdomain = {
	.dom_family = AF_TIPC,
	.dom_name = "tipc",
	.dom_init = tipc_init_domain,
	.dom_protosw = tipcsw,
	.dom_protoswNPROTOSW = &tipcsw[sizeof(tipcsw)/sizeof(tipcsw[0])],
};


/*
 * Anything that resmgr_attach()'s to tcpip_connect_funcs (comes in
 * via af_open()) needs to set (struct msg_open_info *)->path_type.
 */
static struct msg_open_info tipc_info = {
    {0}, 0, AF_TIPC, PATH_TYPE_SOCKET
};



/**
 * tipc_socket_init - initialize TIPC socket interface
 * 
 * Returns 0 on success, errno otherwise
 */
extern void domain_attach(struct domain *dp);
int tipc_socket_init(void)
{
	/* stk_ctl is a global in the iopkt stack */
	extern struct nw_stk_ctl stk_ctl;
	struct nw_stk_ctl *sctlp = &stk_ctl;
	char workpath[16];
    resmgr_attr_t resmgr_attr;
	if (sockets_enabled)
		return EINVAL;

	domain_attach (&tipcdomain);
	iofunc_attr_init(&tipc_info.attr, S_IFSOCK | 0666, 0, 0);

	sprintf (workpath, "/dev/socket/%d", AF_TIPC);
    memset(&resmgr_attr, 0, sizeof(resmgr_attr));
    resmgr_attr.msg_max_size = TIPC_MAX_USER_MSG_SIZE;
	tipc_info.path_id = resmgr_attach(sctlp->dpp, &resmgr_attr/*NULL*/, workpath,
									  _FTYPE_SOCKET, _RESMGR_FLAG_SELF,
									  &tcpip_connect_funcs, NULL, &tipc_info.attr);
	if (tipc_info.path_id == -1) {
	    return errno;
	}

	//PRINT_DBG("%s: tipc sockets initialized\n", __FUNCTION__);
	return EOK;
}

/**
 * tipc_socket_stop - stop TIPC socket interface
 */

void tipc_socket_stop(void)
{
	if (!sockets_enabled)
		return;
}


#endif
